# -*- coding: utf-8; Mode: Python; indent-tabs-mode: nil; tab-width: 4 -*-

# Copyright (C) 2014 nfschina.
# Written by Yuanzhe <yuanzhe@iscas.ac.cn>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import print_function

import locale
import os
import re
import sys

import debconf

from ubiquity import keyboard_names, i18n, misc, osextras, plugin

import ubiquity.tz

NAME = 'protocol'
AFTER = None
WEIGHT = 10


# TODO: This cannot be a non-debconf plugin after all as OEMs may want to
# preseed the 'install updates' and 'install non-free software' options. So?
# Just db_get them. No need for any other overhead, surely. Actually, you
# need the dbfilter for that get.

class PageBase(plugin.PluginUI):

    def __init__(self):
        self.language_choice_map = None
        self.keyboard_choices = None
        self.keyboard_variant_choices = None

    def set_language_choices(self, unused_choices, choice_map):
        """Called with language choices and a map to localised names."""
        self.language_choice_map = dict(choice_map)

    def set_language(self, language):
        """Set the current selected language."""
        pass

    def get_language(self):
        """Get the current selected language."""
        return 'C'

    def set_keyboard_choices(self, choices):
        self.keyboard_choices = choices

    def set_keyboard(self, layout):
        pass

    def get_keyboard(self):
        return ""

    def set_keyboard_variant_choices(self, choices):
        self.keyboard_variant_choices = choices

    def set_keyboard_variant(self, variant):
        pass

    def get_keyboard_variant(self):
        return ""


class PageGtk(PageBase):

    plugin_title = 'ubiquity/text/protocol_heading_label'
    help_title = 'ubiquity/help/title'
    help_filename = "basic_settings"

    def __init__(self, controller, *args, **kwargs):
        PageBase.__init__(self)
        self.controller = controller
        self.current_layout = None
        self.keyboard_layout_timeout_id = 0
        self.keyboard_variant_timeout_id = 0
        self.avalidZone = [["Asia/Beijing", "ubiquity/text/bj_time"],
            ['Asia/Hong_Kong', "ubiquity/text/hk_time"],
            ['Asia/Urumqi', "ubiquity/text/urumqi_time"],
            ['Asia/Taipei', "ubiquity/text/taiwan_time"]]

        self.supportLang = ["中文(简体)", "English"]

        self.keyboardTrans = {"Chinese": "ubiquity/text/kb_ch",
            'Chinese - Tibetan': "ubiquity/text/kb_ch_tibetan1",
            'Chinese - Tibetan (with ASCII numerals)': "ubiquity/text/kb_ch_tibetan2",
            'Chinese - Uyghur': "ubiquity/text/kb_ch_uyghur",
            '其他': "ubiquity/text/kb_other",
            'Other': "ubiquity/text/kb_other"}

        self.tzdb = ubiquity.tz.Database()
        self.timezone = None
        self.online = False
        self.multiple = False

        from gi.repository import Gtk, GObject
        builder = Gtk.Builder()
        self.controller.add_builder(builder)
        builder.add_from_file(os.path.join(
            os.environ['UBIQUITY_GLADE'], 'stepProtocol.ui'))
        builder.connect_signals(self)
        self.page = builder.get_object('stepProtocol')
        self.protocol_label = builder.get_object('protocol_label')
        self.agree = builder.get_object('agree')
        self.plugin_widgets = self.page
        self.frontend = self.controller._wizard

        self.protocol_short_box = builder.get_object("protocol_short_box")
        self.protocol_hypertext = builder.get_object("protocol_hypertext")
        self.lang_label = builder.get_object("lang_label")
        self.keyboard_label = builder.get_object("keyboard_label")
        self.timezone_label = builder.get_object("timezone_label")

        self.lang_selector = builder.get_object('lang_selector')

        self.keyboardlayout_selector = builder.get_object('keyboardlayout_selector')
        self.keyboardvarient_selector = builder.get_object('keyboardvarient_selector')

        self.timezone_selector = builder.get_object('timezone_selector')

        self.protocol_detail_dialog = builder.get_object("protocol_detail_dialog")
        self.protocol_confirm = builder.get_object("protocol_confirm")

        self.lang_model = Gtk.ListStore.new([GObject.TYPE_STRING])
        self.keyboardlayout_model = Gtk.ListStore.new([GObject.TYPE_STRING])
        self.keyboardvarient_model = Gtk.ListStore.new([GObject.TYPE_STRING, GObject.TYPE_STRING])

        self.timezone_model = Gtk.ListStore.new([GObject.TYPE_STRING, GObject.TYPE_STRING])

        self.lang_selector.set_model(self.lang_model)
        self.keyboardlayout_selector.set_model(self.keyboardlayout_model)
        self.keyboardvarient_selector.set_model(self.keyboardvarient_model)
        self.timezone_selector.set_model(self.timezone_model)

        # self.plugin_translate('')
    def init_timezone_selector(self):
        current_timezone = self.get_timezone()
        self.timezone_model.clear()
        for item in self.avalidZone:
            self.timezone_model.append([self.controller.get_string(item[1]), item[0]])
        self.set_timezone(current_timezone)

    def init_keyboard_variant_choises(self, choices=None):
        if choices is None:
            return
        current_kb_var = self.get_keyboard_variant()
        self.keyboardvarient_model.clear()
        for v in sorted(choices):
            if self.keyboardTrans.__contains__(v):
                self.keyboardvarient_model.append([self.controller.get_string(self.keyboardTrans[v]), v])
            else:
                self.keyboardvarient_model.append([v, v])
        if current_kb_var is not None:
            self.set_keyboard_variant(current_kb_var)

    def plugin_set_online_state(self, state):
        self.online = state

    def language_changed(self, widget):
        lang = self.get_language()
        if lang is not None and '' != lang:
            self.controller.translate(lang=lang, just_me=False)
            self.reload_eula(lang)
            self.init_keyboard_variant_choises(self.keyboard_variant_choices)
            self.init_timezone_selector()

    def on_show_protocol_details(self, widget, event):
        self.protocol_detail_dialog.show()
        self.protocol_detail_dialog.run()
        self.protocol_detail_dialog.hide()
        return False

    def keyboardvarient_changed(self, widget):
            # Let's not call this every time the user presses a key.
        from gi.repository import GLib
        if self.keyboard_variant_timeout_id:
            GLib.source_remove(self.keyboard_variant_timeout_id)
        self.keyboard_variant_timeout_id = GLib.timeout_add(
            600, self.keyboard_variant_timeout)

    def timezone_changed(self, widget):
        iterator = self.timezone_selector.get_active_iter()
        value = None
        if iterator is not None:
            value = misc.utf8(self.timezone_model.get_value(iterator, 1))

        self.timezone = value

    def keyboard_variant_timeout(self, *args):
        layout = self.get_keyboard()
        variant = self.get_keyboard_variant()
        if layout is not None and variant is not None:
            self.controller.dbfilter.apply_keyboard(layout, variant)
        return False

    def keyboardlayout_changed(self, widget):
            # Let's not call this every time the user presses a key.
        from gi.repository import GLib
        if self.keyboard_layout_timeout_id:
            GLib.source_remove(self.keyboard_layout_timeout_id)
        self.keyboard_layout_timeout_id = GLib.timeout_add(
            600, self.keyboard_layout_timeout)

    def keyboard_layout_timeout(self, *args):
        layout = self.get_keyboard()
        if layout is not None and layout != self.current_layout:
            self.current_layout = layout
            self.controller.dbfilter.change_layout(layout)
        return False

    def on_agree_clicked(self, widget):
        if widget.get_active():
            self.frontend.next.set_sensitive(True)
        else:
            self.frontend.next.set_sensitive(False)

    def reload_eula(self, lang):
        if lang.startswith('zh_CN'):
            text = open('/etc/cdos/eula.zh_CN').read()
        else:
            text = open('/etc/cdos/eula.en_US').read()
        self.lang_label.set_label(self.controller.get_string('ubiquity/text/choose_language'))
        self.keyboard_label.set_label(self.controller.get_string('ubiquity/text/choose_keyboard'))
        self.timezone_label.set_label(self.controller.get_string('ubiquity/text/choose_timezone'))
        self.protocol_label.set_markup(text)
        self.protocol_hypertext.set_markup("<u><span foreground=\"#0000ff\" font_desc=\"12\">%s</span></u>" % self.controller.get_string('ubiquity/text/protocol_short_title'))
        self.protocol_detail_dialog.set_title(self.controller.get_string('ubiquity/text/protocol_short_title'))
        self.protocol_confirm.set_label(self.controller.get_string("ubiquity/imported/ok"))
        self.agree.set_label(self.controller.get_string('ubiquity/text/agree_protocol'))

    def plugin_on_help_clicked(self):
        self.show_help(self.controller.get_string(self.help_title), self.help_filename)

    def set_language_choices(self, choices, choice_map):
        PageBase.set_language_choices(self, choices, choice_map)
        self.lang_model.clear()
        for choice in self.supportLang:
            self.lang_model.append([choice])

    def set_language(self, language):
        iterator = self.lang_model.iter_children(None)
        while iterator is not None:
            if misc.utf8(self.lang_model.get_value(iterator, 0)) == language:
                self.lang_selector.set_active_iter(iterator)
                break
            iterator = self.lang_model.iter_next(iterator)

    def get_language(self):
        # Support both iconview and treeview
        iterator = self.lang_selector.get_active_iter()
        if iterator is None:
            return None
        else:
            value = misc.utf8(self.lang_model.get_value(iterator, 0))
            return self.language_choice_map[value][1]

    def set_keyboard_choices(self, choices):
        # Sort the choices including these with accents
        import locale
        try:
            # Try with the current locale (will fail if not generated
            # pre-install)
            locale.setlocale(locale.LC_ALL, '')
        except locale.Error:
            try:
                if os.path.exists("/etc/locale.gen"):
                    # Try with the locale generated by casper
                    with open("/etc/locale.gen", "r") as locale_gen:
                        locale.setlocale(
                            locale.LC_ALL, locale_gen.read().split()[0])
                else:
                    # Try a default, usually working locale
                    locale.setlocale(locale.LC_ALL, 'C.UTF-8')
                choices.sort(key=lambda c: locale.strxfrm(c.encode('utf-8')))
            except:
                # Let's avoid crashing when we get a bad locale
                choices.sort()
        PageBase.set_keyboard_choices(choices)
        self.keyboardlayout_model.clear()
        for v in choices:
            self.keyboardlayout_model.append([v])

        if self.current_layout is not None:
            self.set_keyboard(self.current_layout)
        if len(choices) > 0:
            self.keyboardlayout_selector.show()
        else:
            self.keyboardlayout_selector.hide()

    def set_keyboard(self, layout):
        self.current_layout = layout
        iterator = self.keyboardlayout_model.iter_children(None)
        while iterator is not None:
            if misc.utf8(self.keyboardlayout_model.get_value(iterator, 0)) == layout:
                self.keyboardlayout_selector.set_active_iter(iterator)
                break
            iterator = self.keyboardlayout_model.iter_next(iterator)

    def get_keyboard(self):
        iterator = self.keyboardlayout_selector.get_active_iter()
        if iterator is None:
            return None
        else:
            return misc.utf8(self.keyboardlayout_model.get_value(iterator, 0))

    def set_keyboard_variant_choices(self, choices):
        PageBase.set_keyboard_variant_choices(self, choices)
        self.init_keyboard_variant_choises(choices)

    def set_keyboard_variant(self, variant):
        iterator = self.keyboardvarient_model.iter_children(None)
        while iterator is not None:
            if misc.utf8(self.keyboardvarient_model.get_value(iterator, 1)) == variant:
                self.keyboardvarient_selector.set_active_iter(iterator)
                break
            iterator = self.keyboardvarient_model.iter_next(iterator)

    def get_keyboard_variant(self):
        iterator = self.keyboardvarient_selector.get_active_iter()
        if iterator is None:
            return None
        else:
            return misc.utf8(self.keyboardvarient_model.get_value(iterator, 1))

    def set_timezone(self, timezone):
        self.timezone = timezone
        if timezone is None:
            self.timezone_selector.set_active(0)
        else:
            iterator = self.timezone_model.iter_children(None)
            while iterator is not None:
                if misc.utf8(self.timezone_model.get_value(iterator, 1)) == timezone:
                    self.timezone_selector.set_active_iter(iterator)
                    break
                iterator = self.timezone_model.iter_next(iterator)

    def get_timezone(self):
        iterator = self.timezone_selector.get_active_iter()
        if iterator is None:
            return None
        else:
            return misc.utf8(self.timezone_model.get_value(iterator, 1))


class Page(plugin.Plugin):
    def prepare(self, unfiltered=False):
        fullCommand = ['sh', '-c']
        commandstr = ""
        fullQuestion = []
        fullenviron = {}

        self.ui.controller.translate()
        self.ui.reload_eula(self.ui.controller._wizard.locale)
        self.ui.init_timezone_selector()

        self.tzdb = ubiquity.tz.Database()

        self.language_question = None
        self.initial_language = None
        self.db.fset('localechooser/languagelist', 'seen', 'false')
        with misc.raised_privileges():
            osextras.unlink_force('/var/lib/localechooser/preseeded')
            osextras.unlink_force('/var/lib/localechooser/langlevel')

        localechooser_script = '/usr/lib/ubiquity/localechooser/localechooser'
        if ('UBIQUITY_FRONTEND' in os.environ and
                os.environ['UBIQUITY_FRONTEND'] == 'debconf_ui'):
            localechooser_script += '-debconf'

        fullQuestion += ['localechooser/languagelist']

        commandstr += localechooser_script

        if not self.ui.online:
            self.preseed('tzsetup/geoip_server', '')
            self.preseed('clock-setup/ntp', 'false')

        self.multiple = False

        self.db.fset('time/zone', 'seen', 'false')
        cc = self.db.get('debian-installer/country')
        try:
            self.db.get('tzsetup/country/%s' % cc)
            # ... and if that succeeded:
            self.multiple = True
        except debconf.DebconfError:
            pass
        self.preseed('tzsetup/selected', 'false')
        # fullQuestion += ['^time/zone$', '^tzsetup/detected$', 'CAPB', 'PROGRESS']
        # commandstr += ' && /usr/share/ubiquity/clock-setup'

        self.preseed('console-setup/ask_detect', 'false')

        # We need to get rid of /etc/default/keyboard, or console-setup will
        # think it's already configured and behave differently. Try to save
        # the old file for interest's sake, but it's not a big deal if we
        # can't.
        with misc.raised_privileges():
            osextras.unlink_force('/etc/default/keyboard.pre-ubiquity')
            try:
                os.rename('/etc/default/keyboard',
                          '/etc/default/keyboard.pre-ubiquity')
            except OSError:
                osextras.unlink_force('/etc/default/keyboard')
        # Make sure debconf doesn't do anything with crazy "preseeded"
        # answers to these questions. If you want to preseed these, use the
        # *code variants.
        self.db.fset('keyboard-configuration/layout', 'seen', 'false')
        self.db.fset('keyboard-configuration/variant', 'seen', 'false')
        self.db.fset('keyboard-configuration/model', 'seen', 'false')
        self.db.fset('console-setup/codeset47', 'seen', 'false')

        # Roughly taken from console-setup's config.proto:
        di_locale = self.db.get('debian-installer/locale')
        l = di_locale.rsplit('.', 1)[0]
        if not keyboard_names.has_language(l):
            self.debug("No keyboard layout translations for locale '%s'" % l)
            l = l.rsplit('_', 1)[0]
        if not keyboard_names.has_language(l):
            self.debug("No keyboard layout translations for locale '%s'" % l)
            l = 'C'
        self._locale = l

        self.has_variants = False

        # Technically we should provide a version as the second argument,
        # but that isn't currently needed and it would require querying
        # apt/dpkg for the current version, which would be slow, so we don't
        # bother for now.
        commandstr += ' && /usr/lib/ubiquity/console-setup/keyboard-configuration.postinst configure'

        fullQuestion += [
            '^keyboard-configuration/layout',
            '^keyboard-configuration/variant',
            '^keyboard-configuration/model',
            '^keyboard-configuration/altgr$',
            '^keyboard-configuration/unsupported_',
        ]

        fullenviron = {
            'OVERRIDE_ALLOW_PRESEEDING': '1',
            'OVERRIDE_USE_DEBCONF_LOCALE': '1',
            'LC_ALL': di_locale,
            'PATH': '/usr/lib/ubiquity/localechooser:/usr/lib/ubiquity/console-setup/:/usr/share/ubiquity:' + os.environ['PATH'],
        }

        if ('UBIQUITY_FRONTEND' in os.environ and
                os.environ['UBIQUITY_FRONTEND'] == "debconf_ui"):
            fullenviron['TERM_FRAMEBUFFER'] = '1'
        else:
            fullenviron['OVERRIDE_SHOW_ALL_LANGUAGES'] = '1'

        commandstr += ' && /usr/share/ubiquity/simple-plugins protocol'
        fullQuestion += ['ubiquity/agree_protocol']
        self.ui.frontend.next.set_sensitive(False)

        fullCommand.append(commandstr)

        return fullCommand, fullQuestion, fullenviron

    def run(self, priority, question):
        if question == 'localechooser/languagelist':
            self.language_question = question
            if self.initial_language is None:
                self.initial_language = self.db.get(question)
            current_language_index = self.value_index(question)
            only_installable = misc.create_bool(
                self.db.get('ubiquity/only-show-installable-languages'))

            current_language, sorted_choices, language_display_map = \
                i18n.get_languages(current_language_index, only_installable)

            self.ui.set_language_choices(sorted_choices,
                                         language_display_map)

            self.ui.set_language(current_language)
            return True
        elif (self.db.get('ubiquity/skip_eula') == 'true'):
            return True
        elif question == 'keyboard-configuration/layout':
            # TODO cjwatson 2006-09-07: no keyboard-configuration support
            # for layout choice translation yet
            self.ui.set_keyboard_choices(
                self.choices_untranslated(question))
            self.ui.set_keyboard(misc.utf8(self.db.get(question)))
            # Reset these in case we just backed up from the variant
            # question.
            self.store_defaults(True)
            self.has_variants = False
            self.succeeded = True
            return True
        elif question in ('keyboard-configuration/variant',
                          'keyboard-configuration/altgr'):
            if question == 'keyboard-configuration/altgr':
                if self.has_variants:
                    return True
                else:
                    # If there's only one variant, it is always the same as
                    # the layout name.
                    single_variant = misc.utf8(self.db.get(
                        'keyboard-configuration/layout'))
                    self.ui.set_keyboard_variant_choices([single_variant])
                    self.ui.set_keyboard_variant(single_variant)
            else:
                # TODO cjwatson 2006-10-02: no keyboard-configuration
                # support for variant choice translation yet
                self.has_variants = True
                self.ui.set_keyboard_variant_choices(
                    self.choices_untranslated(question))
                self.ui.set_keyboard_variant(misc.utf8(self.db.get(question)))
            # keyboard-configuration preseeding is special, and needs to be
            # checked by hand. The seen flag on
            # keyboard-configuration/layout is used internally by
            # keyboard-configuration, so we can't just force it to true.
            return True
        elif question == 'keyboard-configuration/model':
            # Backing up from the variant question inconveniently goes back
            # to the model question.  Catch this and go forward again so
            # that we can reach the layout question.
            return True
        elif question.startswith('keyboard-configuration/unsupported_'):
            response = self.frontend.question_dialog(
                self.description(question),
                self.extended_description(question),
                ('ubiquity/imported/yes', 'ubiquity/imported/no'))
            if response == 'ubiquity/imported/yes':
                self.preseed(question, 'true')
            else:
                self.preseed(question, 'false')
            return True
        elif question == 'tzsetup/detected' or question == 'time/zone':
            zone = None
            if question == 'tzsetup/detected':
                zone = self.db.get('time/zone')
                self.ui.set_timezone(zone)
            elif question == 'time/zone':
                if self.multiple:
                    # Work around a debconf bug: REGISTER does not appear to
                    # give a newly-registered question the same default as the
                    # question associated with its template, unless we also
                    # RESET it.
                    self.db.reset(question)
                zone = self.db.get(question)
                # Some countries don't have a default zone, so just pick the
                # first choice in the list.
                #  if not zone:
                #      choices_c = self.choices_untranslated(question)
                #      if choices_c:
                #          zone = choices_c[0]
                #  self.ui.set_timezone(zone)
            self.ui.set_timezone(zone)
            if not zone or zone == 'Asia/Shanghai' or zone == 'Asia/Beijing':
                self.ui.timezone_selector.set_active(0)
            elif zone == 'Asia/Hong_Kong':
                self.ui.timezone_selector.set_active(1)
            elif zone == 'Asia/Urumqi':
                self.ui.timezone_selector.set_active(2)
            elif zone == 'Asia/Taipei':
                self.ui.timezone_selector.set_active(3)
            else:
                self.ui.timezone_selector.set_active(0)
            return True
        else:
            return plugin.Plugin.run(self, priority, question)

    def change_layout(self, layout):
        self.preseed('keyboard-configuration/layout', layout)
        self.store_defaults(False)
        # Back up in order to get keyboard-configuration to recalculate the
        # list of possible variants.
        self.succeeded = False
        self.exit_ui_loops()

    def ok_handler(self):
        self.lang_ok_handler()
        self.keyboard_ok_handler()
        self.protocol_ok_handler()
        self.timezone_ok_handler()
        return plugin.Plugin.ok_handler(self)

    def lang_ok_handler(self):
        if self.language_question is not None:
            new_language = self.ui.get_language()
            new_locale = new_language + ".UTF-8"
            self.preseed(self.language_question, new_language)

            os.environ['LANG'] = new_locale
            os.environ['LANGUAGE'] = new_locale
            try:
                locale.setlocale(locale.LC_ALL, '')
            except locale.Error as e:
                print('locale.setlocale failed: %s (LANG=%s)' % (e, new_locale),
                      file=sys.stderr)
                new_locale = "C.UTF-8"
                os.environ['LANG'] = new_locale
                os.environ['LANGUAGE'] = new_locale
                try:
                    locale.setlocale(locale.LC_ALL, '')
                except locale.Error as e:
                    print('locale.setlocale failed: %s (LANG=%s)' % (e, new_locale),
                          file=sys.stderr)

            if (self.initial_language is None or
                    self.initial_language != new_language):
                self.db.reset('debian-installer/country')

    def protocol_ok_handler(self):
        self.preseed_bool('ubiquity/agree_protocol', True)

    def keyboard_ok_handler(self):
        variant = self.ui.get_keyboard_variant()
        if variant is not None:
            self.preseed('keyboard-configuration/variant', variant)

        if hasattr(self.ui, "cancel_timeouts"):
            self.ui.cancel_timeouts()

    def timezone_ok_handler(self):
        zone = self.ui.get_timezone()
        if zone is None:
            zone = self.db.get('time/zone')
        else:
            self.preseed('time/zone', zone)
        self.set_di_country(zone)

    def set_di_country(self, zone):
        location = self.tzdb.get_loc(zone)
        if location:
            self.preseed('debian-installer/country', location.country)

    def get_locale(self):
        return self._locale

    def apply_keyboard(self, layout_name, variant_name):
        model = self.db.get('keyboard-configuration/modelcode')

        l = self.get_locale()
        try:
            layout = keyboard_names.layout_id(l, layout_name)
        except KeyError:
            self.debug("Unknown keyboard layout '%s'" % layout_name)
            return

        if not keyboard_names.has_variants(l, layout):
            self.debug("No known variants for layout '%s'" % layout)
            variant = ''
        else:
            try:
                variant = keyboard_names.variant_id(
                    l, layout, variant_name)
            except KeyError:
                self.debug("Unknown keyboard variant '%s' for layout '%s'" %
                           (variant_name, layout_name))
                return

        (model, layout, variant, options) = \
            self.adjust_keyboard(model, layout, variant, [])
        self.debug("Setting keyboard layout: %s %s %s %s" %
                   (model, layout, variant, options))
        self.apply_real_keyboard(model, layout, variant, options)

    def apply_real_keyboard(self, model, layout, variant, options):
        args = []
        if model is not None and model != '':
            args.extend(("-model", model))
        args.extend(("-layout", layout))
        if variant != '':
            args.extend(("-variant", variant))
        args.extend(("-option", ""))
        for option in options:
            args.extend(("-option", option))
        misc.execute("setxkbmap", *args)

    def store_defaults(self, store):
        self.preseed_bool(
            'keyboard-configuration/store_defaults_in_debconf_db', store,
            seen=False)

    def adjust_keyboard(self, model, layout, variant, options):
        """Apply any necessary tweaks to the supplied model, layout, variant,
        and options."""

        if layout in ('am', 'ara', 'ben', 'bd', 'bg', 'bt', 'by', 'deva', 'ge',
                      'gh', 'gr', 'guj', 'guru', 'il', 'in', 'ir', 'iku',
                      'kan', 'kh', 'kz', 'la', 'lao', 'lk', 'mk', 'mm', 'mn',
                      'mv', 'mal', 'ori', 'pk', 'ru', 'scc', 'sy', 'syr',
                      'tel', 'th', 'tj', 'tam', 'ua', 'uz'):
            latin = False
            real_layout = 'us,%s' % layout
        elif layout == 'jp':
            if variant in ('106', 'common', 'OADG109A', 'nicola_f_bs', ''):
                latin = True
                real_layout = layout
            else:
                latin = False
                real_layout = 'jp,jp'
        elif layout == 'lt':
            latin = False
            real_layout = 'lt,lt'
        elif layout == 'me':
            if variant == 'basic' or variant.startswith('latin'):
                latin = True
                real_layout = layout
            else:
                latin = False
                real_layout = 'me,me'
        elif layout == 'rs':
            if variant == 'basic' or variant.startswith('latin'):
                latin = True
                real_layout = layout
            else:
                latin = False
                real_layout = 'rs,rs'
        else:
            latin = True
            real_layout = layout

        if latin:
            real_variant = variant
        elif real_layout == 'jp,jp':
            real_variant = '106,%s' % variant
        elif real_layout == 'lt,lt':
            if variant == 'us':
                real_variant = 'us,'
            else:
                real_variant = '%s,us' % variant
        elif real_layout == 'me,me':
            if variant == 'cyrillicyz':
                real_variant = 'latinyz,%s' % variant
            elif variant == 'cyrillicalternatequotes':
                real_variant = 'latinalternatequotes,%s' % variant
            else:
                real_variant = 'basic,%s' % variant
        elif real_layout == 'rs,rs':
            if variant == 'yz':
                real_variant = 'latinyz,%s' % variant
            elif variant == 'alternatequotes':
                real_variant = 'latinalternatequotes,%s' % variant
            else:
                real_variant = 'latin,%s' % variant
        else:
            real_variant = ',%s' % variant

        real_options = [opt for opt in options if not opt.startswith('lv3:')]
        if not latin:
            toggle = re.compile(r'^grp:.*toggle$')
            real_options = [
                opt for opt in real_options if not toggle.match(opt)]
            # TODO cjwatson 2006-09-07: honour crazy preseeding; probably
            # not quite right, especially for Apples which may need a level
            # 3 shift
            real_options.append('grp:alt_shift_toggle')
        if layout != 'us':
            real_options.append('lv3:ralt_switch')

        real_model = model
        if model == 'pc105':
            if real_layout == 'br':
                real_model = 'abnt2'
            elif real_layout == 'jp':
                real_model = 'jp106'

        return (real_model, real_layout, real_variant, real_options)


class Install(plugin.InstallPlugin):
    def prepare(self, unfiltered=False):
        fullCommand = ['sh', '-c']
        questions = []

        commandstr = ""
        if 'UBIQUITY_OEM_USER_CONFIG' in os.environ:
            commandstr += '/usr/share/ubiquity/localechooser-apply'
        else:
            commandstr += '/usr/lib/ubiquity/localechooser/post-base-installer && /usr/lib/ubiquity/localechooser/finish-install'

        commandstr += ' && /usr/lib/ubiquity/tzsetup/post-base-installer && /usr/share/ubiquity/clock-setup-apply'

        commandstr += ' && /usr/share/ubiquity/console-setup-apply'

        fullCommand.append(commandstr)
        return fullCommand, questions

    def install(self, target, progress, *args, **kwargs):
        langRet = self.lang_install(target, progress, *args, **kwargs)
        keyboardRet = self.keyboard_install(target, progress, *args, **kwargs)
        return langRet and keyboardRet

    def _pam_env_lines(self, fd):
        """Yield a sequence of logical lines from a PAM environment file."""
        buf = ""
        for line in fd:
            line = line.lstrip()
            if line and line[0] != "#":
                if "#" in line:
                    line = line[:line.index("#")]
                for i in range(len(line) - 1, -1, -1):
                    if line[i].isspace():
                        break
                if line[i] == "\\":
                    # Continuation line.
                    buf = line[:i]
                else:
                    buf += line
                    yield buf
                    buf = ""
        if buf:
            yield buf

    def _pam_env_parse_file(self, path):
        """Parse a PAM environment file just as pam_env does.

        We use this for reading /etc/default/locale after configuring it.
        """
        try:
            with open(path) as fd:
                for line in self._pam_env_lines(fd):
                    line = line.lstrip()
                    if not line or line[0] == "#":
                        continue
                    if line.startswith("export "):
                        line = line[7:]
                    line = re.sub(r"[#\n].*", "", line)
                    if not re.match(r"^[A-Z_]+=", line):
                        continue
                    # pam_env's handling of quoting is crazy, but in this
                    # case it's better to imitate it than to fix it.
                    key, value = line.split("=", 1)
                    if value.startswith('"') or value.startswith("'"):
                        value = re.sub(r"[\"'](.|$)", r"\1", value[1:])
                    yield key, value
        except IOError:
            pass

    def lang_install(self, target, progress, *args, **kwargs):
        progress.info('ubiquity/install/locales')
        rv = plugin.InstallPlugin.install(
            self, target, progress, *args, **kwargs)
        if not rv:
            locale_file = "/etc/default/locale"
            if 'UBIQUITY_OEM_USER_CONFIG' not in os.environ:
                locale_file = "/target%s" % locale_file
            for key, value in self._pam_env_parse_file(locale_file):
                if key in ("LANG", "LANGUAGE") or key.startswith("LC_"):
                    os.environ[key] = value
            # Start using the newly-generated locale, if possible.
            try:
                locale.setlocale(locale.LC_ALL, '')
            except locale.Error:
                pass
        return rv

    def keyboard_install(self, target, progress, *args, **kwargs):
        progress.info('ubiquity/install/keyboard')
        return plugin.InstallPlugin.install(
            self, target, progress, *args, **kwargs)

    def timezone_install(self, target, progress, *args, **kwargs):
        progress.info('ubiquity/install/timezone')
        return plugin.InstallPlugin.install(
            self, target, progress, *args, **kwargs)
